"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var TargetLanguage_1 = require("../TargetLanguage");
var Type_1 = require("../Type");
var Naming_1 = require("../Naming");
var Source_1 = require("../Source");
var Strings_1 = require("../Strings");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var RendererOptions_1 = require("../RendererOptions");
var unicode = require("unicode-properties");
var ObjectiveCTargetLanguage = /** @class */ (function (_super) {
    __extends(ObjectiveCTargetLanguage, _super);
    function ObjectiveCTargetLanguage() {
        var _this = _super.call(this, "Objective-C", ["objc", "objective-c", "objectivec"], "m") || this;
        _this._featuresOption = new RendererOptions_1.EnumOption("features", "Interface and implementation", [
            ["all", { interface: true, implementation: true }],
            ["interface", { interface: true, implementation: false }],
            ["implementation", { interface: false, implementation: true }]
        ]);
        _this._justTypesOption = new RendererOptions_1.BooleanOption("just-types", "Plain types only", false);
        _this._classPrefixOption = new RendererOptions_1.StringOption("class-prefix", "Class prefix", "PREFIX", "QT");
        _this._extraCommentsOption = new RendererOptions_1.BooleanOption("extra-comments", "Extra comments", false);
        _this.setOptions([
            _this._justTypesOption,
            _this._classPrefixOption,
            _this._featuresOption,
            _this._extraCommentsOption
        ]);
        return _this;
    }
    Object.defineProperty(ObjectiveCTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return ObjectiveCRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return ObjectiveCTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = ObjectiveCTargetLanguage;
function typeNameStyle(prefix, original) {
    var words = Strings_1.splitIntoWords(original);
    var result = Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
    // Take care not to doubly-prefix type names
    return lodash_1.startsWith(result, prefix) ? result : prefix + result;
}
function propertyNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.allLowerWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allLowerWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
}
var keywords = [
    "_Bool",
    "_Complex",
    "_Imaginary",
    "atomic",
    "auto",
    "break",
    "case ",
    "char",
    "const",
    "continue",
    "default",
    "do",
    "double",
    "else",
    "enum",
    "extern",
    "false",
    "float ",
    "for ",
    "goto",
    "if",
    "inline",
    "int",
    "long",
    "nil",
    "nonatomic",
    "register",
    "restrict",
    "retain",
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "struct",
    "switch",
    "typedef",
    "true",
    "union",
    "unsigned",
    "void",
    "volatile",
    "while"
];
var propertySafeKeywords = [
    "BOOL",
    "Class",
    "bycopy",
    "byref",
    "decimal",
    "id",
    "IMP",
    "in",
    "inout",
    "NO",
    "NULL",
    "oneway",
    "out",
    "Protocol",
    "SEL",
    "self",
    "super",
    "YES"
];
function isStartCharacter(utf16Unit) {
    return unicode.isAlphabetic(utf16Unit) || utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    return lodash_1.includes(["Nd", "Pc", "Mn", "Mc"], category) || isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function isAnyOrNull(t) {
    return t.kind === "any" || t.kind === "null";
}
var ObjectiveCRenderer = /** @class */ (function (_super) {
    __extends(ObjectiveCRenderer, _super);
    function ObjectiveCRenderer(graph, leadingComments, _justTypes, _classPrefix, _features, _extraComments) {
        var _this = _super.call(this, graph, leadingComments) || this;
        _this._justTypes = _justTypes;
        _this._classPrefix = _classPrefix;
        _this._features = _features;
        _this._extraComments = _extraComments;
        _this.emitBlock = function (line, f) {
            _this.emitLine(line, " {");
            _this.indent(f);
            _this.emitLine("}");
        };
        _this.emitExtraComments = function () {
            var comments = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                comments[_i] = arguments[_i];
            }
            if (!_this._extraComments)
                return;
            for (var _a = 0, comments_1 = comments; _a < comments_1.length; _a++) {
                var comment = comments_1[_a];
                _this.emitLine("// ", comment);
            }
        };
        _this.objcType = function (t) {
            return Type_1.matchType(t, function (_anyType) { return ["id", ""]; }, 
            // For now, we're treating nulls just like any
            function (_nullType) { return ["id", ""]; }, function (_boolType) { return ["NSBoolean", " *"]; }, function (_integerType) { return ["NSNumber", " *"]; }, function (_doubleType) { return ["NSNumber", " *"]; }, function (_stringType) { return ["NSString", " *"]; }, function (arrayType) {
                var itemType = arrayType.items;
                var itemTypeName = _this.objcType(itemType);
                // NSArray<id>* === NSArray*
                if (isAnyOrNull(itemType)) {
                    return ["NSArray", " *"];
                }
                return [["NSArray<", itemTypeName, ">"], " *"];
            }, function (classType) { return [_this.nameForNamedType(classType), " *"]; }, function (mapType) { return [["NSDictionary<NSString *, ", _this.objcType(mapType.values), ">"], " *"]; }, 
            // TODO Support enums
            function (_enumType) { return ["NSString", " *"]; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                return nullable !== null ? _this.objcType(nullable) : ["id", ""];
            });
        };
        _this.jsonType = function (t) {
            return Type_1.matchType(t, function (_anyType) { return ["id", ""]; }, 
            // For now, we're treating nulls just like any
            function (_nullType) { return ["id", ""]; }, function (_boolType) { return ["NSBoolean", " *"]; }, function (_integerType) { return ["NSNumber", " *"]; }, function (_doubleType) { return ["NSNumber", " *"]; }, function (_stringType) { return ["NSString", " *"]; }, function (_arrayType) { return ["NSArray", " *"]; }, function (_classType) { return ["NSDictionary<NSString *, id>", " *"]; }, function (mapType) { return [["NSDictionary<NSString *, ", _this.jsonType(mapType.values), ">"], " *"]; }, 
            // TODO Support enums
            function (_enumType) { return ["NSString", " *"]; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                return nullable !== null ? _this.jsonType(nullable) : ["id", ""];
            });
        };
        _this.fromDynamicExpression = function (t) {
            var dynamic = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                dynamic[_i - 1] = arguments[_i];
            }
            return Type_1.matchType(t, function (_anyType) { return dynamic; }, function (_nullType) { return dynamic; }, function (_boolType) { return dynamic; }, function (_integerType) { return dynamic; }, function (_doubleType) { return dynamic; }, function (_stringType) { return dynamic; }, function (arrayType) { return ["[", dynamic, " map:λ(id x, ", _this.fromDynamicExpression(arrayType.items, "x"), ")]"]; }, function (classType) { return ["[", _this.nameForNamedType(classType), " fromJSONDictionary:", dynamic, "]"]; }, function (mapType) { return ["[", dynamic, " map:λ(id x, ", _this.fromDynamicExpression(mapType.values, "x"), ")]"]; }, 
            // TODO Support enums
            function (_enumType) { return dynamic; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                return nullable !== null ? _this.fromDynamicExpression(nullable, dynamic) : dynamic;
            });
        };
        _this.toDynamicExpression = function (t, dynamic) {
            return Type_1.matchType(t, function (_anyType) { return ["NSNullify(", dynamic, ")"]; }, function (_nullType) { return ["NSNullify(", dynamic, ")"]; }, function (_boolType) { return dynamic; }, function (_integerType) { return dynamic; }, function (_doubleType) { return dynamic; }, function (_stringType) { return dynamic; }, function (arrayType) {
                if (_this.isJSONSafe(arrayType)) {
                    // TODO check each value type
                    return dynamic;
                }
                return ["[", dynamic, " map:λ(id x, ", _this.toDynamicExpression(arrayType.items, "x"), ")]"];
            }, function (_classType) { return ["[", dynamic, " JSONDictionary]"]; }, function (mapType) {
                if (_this.isJSONSafe(mapType)) {
                    // TODO check each value type
                    return dynamic;
                }
                return ["[", dynamic, " map:λ(id x, ", _this.toDynamicExpression(mapType.values, "x"), ")]"];
            }, 
            // TODO Support enums
            function (_enumType) { return dynamic; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable !== null) {
                    if (_this.isJSONSafe(nullable)) {
                        return ["NSNullify(", dynamic, ")"];
                    }
                    else {
                        return [
                            "NSNullMap(",
                            ["NSNullify(", dynamic, ")"],
                            ", λ(id x, ",
                            _this.toDynamicExpression(nullable, "x"),
                            "))"
                        ];
                    }
                }
                else {
                    // TODO support unions
                    return dynamic;
                }
            });
        };
        _this.safePropertyName = function (propertyName) {
            var isKeyword = lodash_1.includes(propertySafeKeywords, _this.sourcelikeToString(propertyName));
            return isKeyword ? ["self.", propertyName] : propertyName;
        };
        _this.emitPropertyAssignment = function (propertyName, _json, propertyType) {
            var key = Strings_1.stringEscape(_json);
            var name = _this.safePropertyName(propertyName);
            Type_1.matchType(propertyType, function (anyType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(anyType, '[dict objectForKey:@"', key, '"]'), ";");
            }, function (nullType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(nullType, '[dict objectForKey:@"', key, '"]'), ";");
            }, function (boolType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(boolType, '[dict boolForKey:@"', key, '"]'), ";");
            }, function (integerType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(integerType, '[dict numberForKey:@"', key, '"]'), ";");
            }, function (doubleType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(doubleType, '[dict numberForKey:@"', key, '"]'), ";");
            }, function (stringType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(stringType, '[dict stringForKey:@"', key, '"]'), ";");
            }, function (arrayType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(arrayType, '[dict arrayForKey:@"', key, '"]'), ";");
            }, function (classType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(classType, '[dict dictionaryForKey:@"', key, '"]'), ";");
            }, function (mapType) {
                var itemType = mapType.values;
                if (itemType.isPrimitive() ||
                    // TODO once enums are supported, this will change
                    itemType.kind === "enum" ||
                    // Before union support, we have a lot of untyped data
                    // This ensures that we don't map over unknown elements
                    isAnyOrNull(itemType)) {
                    // TODO check each value type
                    _this.emitLine(name, ' = [dict dictionaryForKey:@"', key, '"];');
                }
                else {
                    _this.emitLine(name, ' = [[dict dictionaryForKey:@"', key, '"] map:', ["λ(id x, ", _this.fromDynamicExpression(itemType, "x"), ")"], "];");
                }
            }, 
            // TODO Support enums
            function (enumType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(enumType, '[dict stringForKey:@"', key, '"]'), ";");
            }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable !== null) {
                    _this.ensureBlankLine();
                    _this.emitBlock(['if (!NSNullOrNil([dict objectForKey:@"', key, '"]))'], function () {
                        _this.emitPropertyAssignment(propertyName, _json, nullable);
                    });
                    _this.ensureBlankLine();
                }
                else {
                    // TODO This is a union, but for now we just leave it dynamic
                    _this.emitLine(name, " = ", _this.fromDynamicExpression(unionType, '[dict objectForKey:@"', key, '"]'), ";");
                }
                return nullable !== null ? _this.objcType(nullable) : "id";
            });
        };
        _this.emitPrivateClassInterface = function (_, name) {
            _this.emitLine("@interface ", name, " (JSONConversion)");
            _this.emitLine("+ (instancetype)fromJSONDictionary:(NSDictionary *)dict;");
            _this.emitLine("- (NSDictionary *)JSONDictionary;");
            _this.emitLine("@end");
        };
        _this.emitClassInterface = function (t, className) {
            var isTopLevel = _this.topLevels.valueSeq().contains(t);
            _this.emitLine("@interface ", className, " : NSObject");
            _this.forEachClassProperty(t, "none", function (name, _json, propertyType) {
                var attributes = ["nonatomic"];
                // TODO offer a 'readonly' option
                // TODO We must add "copy" if it's NSCopy, otherwise "strong"
                if (propertyType.isNullable) {
                    attributes.push("nullable");
                }
                _this.emitLine("@property ", ["(", attributes.join(", "), ")"], " ", _this.pointerAwareTypeName(propertyType), name, ";");
            });
            if (!_this._justTypes && isTopLevel) {
                if (t.properties.count() > 0)
                    _this.ensureBlankLine();
                _this.emitLine("+ (_Nullable instancetype)fromJSON:(NSString *)json;");
                _this.emitLine("+ (_Nullable instancetype)fromJSON:(NSString *)json encoding:(NSStringEncoding)encoding error:(NSError *_Nullable *)error;");
                _this.emitLine("+ (_Nullable instancetype)fromData:(NSData *)data error:(NSError *_Nullable *)error;");
                _this.emitLine("- (NSString *_Nullable)toJSON;");
                _this.emitLine("- (NSString *_Nullable)toJSON:(NSStringEncoding)encoding error:(NSError *_Nullable *)error;");
                _this.emitLine("- (NSData *_Nullable)toData:(NSError *_Nullable *)error;");
            }
            _this.emitLine("@end");
        };
        // TODO Implement NSCopying
        _this.emitClassImplementation = function (t, className) {
            var isTopLevel = _this.topLevels.valueSeq().contains(t);
            _this.emitLine("@implementation ", className);
            var propertyNames = [];
            _this.forEachClassProperty(t, "none", function (name) {
                if (propertyNames.length !== 0) {
                    propertyNames.push(", ");
                }
                propertyNames.push(name);
                if (propertyNames.length === 5 * 2 - 1) {
                    _this.emitLine("@synthesize ", propertyNames, ";");
                    propertyNames = [];
                }
            });
            if (propertyNames.length !== 0) {
                _this.emitLine("@synthesize ", propertyNames, ";");
            }
            if (!_this._justTypes) {
                _this.ensureBlankLine();
                if (isTopLevel) {
                    _this.emitBlock("+ (_Nullable instancetype)fromData:(NSData *)data error:(NSError *_Nullable *)error", function () {
                        _this.emitLine("return ", className, "FromData(data, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("+ (_Nullable instancetype)fromJSON:(NSString *)json encoding:(NSStringEncoding)encoding error:(NSError *_Nullable *)error;", function () {
                        _this.emitLine("return ", className, "FromJSON(json, encoding, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("+ (_Nullable instancetype)fromJSON:(NSString *)json", function () {
                        _this.emitLine("NSError *error;");
                        _this.emitLine("return ", className, "FromJSON(json, NSUTF8StringEncoding, &error);");
                    });
                }
                _this.ensureBlankLine();
                _this.emitBlock("+ (instancetype)fromJSONDictionary:(NSDictionary *)dict", function () {
                    _this.emitLine("return [[", className, " alloc] initWithJSONDictionary:dict];");
                });
                _this.ensureBlankLine();
                _this.emitBlock("- (instancetype)initWithJSONDictionary:(NSDictionary *)dict", function () {
                    _this.emitBlock("if (self = [super init])", function () {
                        _this.forEachClassProperty(t, "none", _this.emitPropertyAssignment);
                    });
                    _this.emitLine("return self;");
                });
                _this.ensureBlankLine();
                _this.emitBlock("- (NSDictionary *)JSONDictionary", function () {
                    _this.emitLine("return @{");
                    _this.indent(function () {
                        _this.forEachClassProperty(t, "none", function (propertyName, jsonKey, propertyType) {
                            var key = Strings_1.stringEscape(jsonKey);
                            var name = _this.safePropertyName(propertyName);
                            _this.emitLine('@"', key, '": ', _this.toDynamicExpression(propertyType, name), ",");
                        });
                    });
                    _this.emitLine("};");
                });
                if (isTopLevel) {
                    _this.ensureBlankLine();
                    _this.emitBlock("- (NSData *_Nullable)toData:(NSError *_Nullable *)error", function () {
                        _this.emitLine("return ", className, "ToData(self, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("- (NSString *_Nullable)toJSON:(NSStringEncoding)encoding error:(NSError *_Nullable *)error", function () {
                        _this.emitLine("return ", className, "ToJSON(self, encoding, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("- (NSString *_Nullable)toJSON", function () {
                        _this.emitLine("NSError *error;");
                        _this.emitLine("return ", className, "ToJSON(self, NSUTF8StringEncoding, &error);");
                    });
                }
            }
            _this.emitLine("@end");
        };
        _this.emitMark = function (label) {
            _this.ensureBlankLine();
            _this.emitLine("// MARK: " + label);
            _this.ensureBlankLine();
        };
        _this.emitDictionaryAndArrayExtensions = function () {
            _this.emitMultiline("@implementation NSArray (JSONConversion)\n- (NSArray *)map:(id (^)(id element))f {\n    id result = [NSMutableArray arrayWithCapacity:self.count];\n    for (id x in self) [result addObject:f(x)];\n    return result;\n}\n@end\n\n@implementation NSDictionary (JSONConversion)\n- (NSDictionary *)map:(id (^)(id value))f {\n    id result = [NSMutableDictionary dictionaryWithCapacity:self.count];\n    for (id key in self) [result setObject:f([self objectForKey:key]) forKey:key];\n    return result;\n}\n\n- (NSException *)exceptionForKey:(id)key type:(NSString *)type {\n    return [NSException exceptionWithName:@\"TypeException\"\n                                    reason:[NSString stringWithFormat:@\"Expected a %@\", type]\n                                    userInfo:@{ @\"dictionary\":self, @\"key\":key }];\n}\n\n- (NSString *)stringForKey:(NSString *)key {\n    id value = [self objectForKey:key];\n    if ([value isKindOfClass:[NSString class]]) {\n        return value;\n    } else {\n        @throw [self exceptionForKey:key type:@\"string\"];\n    }\n}\n\n- (NSNumber *)numberForKey:(NSString *)key {\n    id value = [self objectForKey:key];\n    // TODO Could this check be more precise?\n    if ([value isKindOfClass:[NSNumber class]]) {\n        return value;\n    } else {\n        @throw [self exceptionForKey:key type:@\"number\"];\n    }\n}\n\n- (NSBoolean *)boolForKey:(NSString *)key {\n    id value = [self objectForKey:key];\n    if ([value isEqual:@(YES)] || [value isEqual:@(NO)]) {\n        return value;\n    } else {\n        @throw [self exceptionForKey:key type:@\"bool\"];\n    }\n}\n\n- (NSArray *)arrayForKey:(NSString *)key {\n    id value = [self objectForKey:key];\n    if ([value isKindOfClass:[NSArray class]]) {\n        return value;\n    } else {\n        @throw [self exceptionForKey:key type:@\"array\"];\n    }\n}\n\n- (NSDictionary *)dictionaryForKey:(NSString *)key {\n    id value = [self objectForKey:key];\n    if ([value isKindOfClass:[NSDictionary class]]) {\n        return value;\n    } else {\n        @throw [self exceptionForKey:key type:@\"dictionary\"];\n    }\n}\n@end");
        };
        return _this;
    }
    Object.defineProperty(ObjectiveCRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return keywords;
        },
        enumerable: true,
        configurable: true
    });
    ObjectiveCRenderer.prototype.forbiddenForClassProperties = function (_c, _classNamed) {
        return { names: [], namespaces: [this.forbiddenWordsNamespace] };
    };
    ObjectiveCRenderer.prototype.forbiddenForEnumCases = function (_e, _enumNamed) {
        return { names: [], namespaces: [this.forbiddenWordsNamespace] };
    };
    ObjectiveCRenderer.prototype.topLevelNameStyle = function (rawName) {
        return Strings_1.camelCase(rawName);
    };
    ObjectiveCRenderer.prototype.makeNamedTypeNamer = function () {
        var _this = this;
        return Naming_1.funPrefixNamer(function (rawName) { return typeNameStyle(_this._classPrefix, rawName); });
    };
    ObjectiveCRenderer.prototype.makeClassPropertyNamer = function () {
        // TODO why is underscore being removed?
        return new Naming_1.Namer(function (s) { return propertyNameStyle(s); }, ["_", "the", "one", "some", "another"]);
    };
    ObjectiveCRenderer.prototype.makeUnionMemberNamer = function () {
        return null;
    };
    ObjectiveCRenderer.prototype.makeEnumCaseNamer = function () {
        return new Naming_1.Namer(propertyNameStyle, []);
    };
    ObjectiveCRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        return type;
    };
    ObjectiveCRenderer.prototype.isJSONSafe = function (t) {
        if (t instanceof Type_1.ClassType) {
            return false;
        }
        else if (t instanceof Type_1.EnumType) {
            return true;
        }
        else if (t instanceof Type_1.ArrayType) {
            return this.isJSONSafe(t.items);
        }
        else if (t instanceof Type_1.MapType) {
            return this.isJSONSafe(t.values);
        }
        else if (t.isPrimitive()) {
            return true;
        }
        else {
            return false;
        }
    };
    ObjectiveCRenderer.prototype.pointerAwareTypeName = function (t) {
        var name = t instanceof Type_1.Type ? this.objcType(t) : t;
        var isPointer = name[1] !== "";
        return isPointer ? name : [name, " "];
    };
    ObjectiveCRenderer.prototype.emitNonClassTopLevelTypedef = function (t, name) {
        var nonPointerTypeName = this.objcType(t)[0];
        this.emitLine("typedef ", nonPointerTypeName, " ", name, ";");
    };
    ObjectiveCRenderer.prototype.topLevelFromDataPrototype = function (name) {
        return [name, " *", name, "FromData(NSData *data, NSError **error)"];
    };
    ObjectiveCRenderer.prototype.topLevelFromJSONPrototype = function (name) {
        return [name, " *", name, "FromJSON(NSString *json, NSStringEncoding encoding, NSError **error)"];
    };
    ObjectiveCRenderer.prototype.topLevelToDataPrototype = function (name) {
        var parameter = this.variableNameForTopLevel(name);
        return ["NSData *", name, "ToData(", name, " *", parameter, ", NSError **error)"];
    };
    ObjectiveCRenderer.prototype.topLevelToJSONPrototype = function (name) {
        var parameter = this.variableNameForTopLevel(name);
        return ["NSString *", name, "ToJSON(", name, " *", parameter, ", NSStringEncoding encoding, NSError **error)"];
    };
    ObjectiveCRenderer.prototype.emitTopLevelFunctionDeclarations = function (_, name) {
        this.emitExtraComments(name);
        this.emitLine(this.topLevelFromDataPrototype(name), ";");
        this.emitLine(this.topLevelFromJSONPrototype(name), ";");
        this.emitLine(this.topLevelToDataPrototype(name), ";");
        this.emitLine(this.topLevelToJSONPrototype(name), ";");
    };
    ObjectiveCRenderer.prototype.emitTopLevelFunctions = function (t, name) {
        var _this = this;
        var parameter = this.variableNameForTopLevel(name);
        this.ensureBlankLine();
        this.emitBlock(this.topLevelFromDataPrototype(name), function () {
            _this.emitLine(_this.pointerAwareTypeName(_this.jsonType(t)), "json = [NSJSONSerialization JSONObjectWithData:data options:NSJSONReadingAllowFragments error:error];");
            _this.emitLine("return *error ? nil : ", _this.fromDynamicExpression(t, "json"), ";");
        });
        this.ensureBlankLine();
        this.emitBlock(this.topLevelFromJSONPrototype(name), function () {
            _this.emitLine("return ", name, "FromData([json dataUsingEncoding:encoding], error);");
        });
        this.ensureBlankLine();
        this.emitBlock(this.topLevelToDataPrototype(name), function () {
            _this.emitLine(_this.pointerAwareTypeName(_this.jsonType(t)), "json = ", _this.toDynamicExpression(t, parameter), ";");
            _this.emitLine("NSData *data = [NSJSONSerialization dataWithJSONObject:json options:kNilOptions error:error];");
            _this.emitLine("return *error ? nil : data;");
        });
        this.ensureBlankLine();
        this.emitBlock(this.topLevelToJSONPrototype(name), function () {
            _this.emitLine("NSData *data = ", name, "ToData(", parameter, ", error);");
            _this.emitLine("return data ? [[NSString alloc] initWithData:data encoding:encoding] : nil;");
        });
    };
    ObjectiveCRenderer.prototype.variableNameForTopLevel = function (name) {
        var _this = this;
        var camelCaseName = Source_1.modifySource(function (serialized) {
            // 1. remove class prefix
            serialized = serialized.substr(_this._classPrefix.length);
            // 2. camel case
            return Strings_1.camelCase(serialized);
        }, name);
        return camelCaseName;
    };
    ObjectiveCRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        if (this.leadingComments !== undefined) {
            this.emitCommentLines("// ", this.leadingComments);
        }
        else if (!this._justTypes && !this._features.interface) {
            this.emitLine("// Remember to import the companion header like:");
            this.emitLine("// ");
            this.emitLine('//   #import "', this._classPrefix, 'Header.h"');
        }
        else if (!this._justTypes) {
            this.emitCommentLines("// ", ["To parse this JSON:", ""]);
            this.emitLine("//   NSError *error;");
            this.forEachTopLevel("none", function (t, topLevelName) {
                var fromJsonExpression = t instanceof Type_1.ClassType
                    ? ["[", topLevelName, " fromJSON:json encoding:NSUTF8Encoding error:&error]"]
                    : [topLevelName, "FromJSON(json, NSUTF8Encoding, &error);"];
                _this.emitLine("//   ", topLevelName, " *", _this.variableNameForTopLevel(topLevelName), " = ", fromJsonExpression);
            });
        }
        this.ensureBlankLine();
        this.emitLine("#import <Foundation/Foundation.h>");
        this.ensureBlankLine();
        if (this._features.interface) {
            if (this._features.implementation) {
                this.emitMark("Types");
            }
            this.ensureBlankLine();
            this.emitExtraComments("This clarifies which properties are JSON booleans");
            this.emitLine("typedef NSNumber NSBoolean;");
            // Emit @class declarations for top-level array+maps and classes
            this.ensureBlankLine();
            this.forEachClass("none", function (_, name) { return _this.emitLine("@class ", name, ";"); });
            this.ensureBlankLine();
            this.ensureBlankLine();
            this.emitLine("NS_ASSUME_NONNULL_BEGIN");
            this.ensureBlankLine();
            // Emit interfaces for top-level array+maps and classes
            this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitNonClassTopLevelTypedef(t, n); }, function (t) { return !(t instanceof Type_1.ClassType); });
            if (!this._justTypes) {
                this.emitMark("Top-level marshalling functions");
                this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitTopLevelFunctionDeclarations(t, n); });
            }
            this.forEachNamedType("leading-and-interposing", this.emitClassInterface, function () { return null; }, function () { return null; });
            this.ensureBlankLine();
            this.emitLine("NS_ASSUME_NONNULL_END");
        }
        if (this._features.implementation) {
            if (this._features.interface) {
                this.emitMark("Implementation");
            }
            if (!this._justTypes) {
                this.emitExtraComments("Shorthand for simple blocks");
                this.emitLine("#define \u03BB(decl, expr) (^(decl) { return (expr); })");
                this.ensureBlankLine();
                this.emitLine("NS_ASSUME_NONNULL_BEGIN");
                this.ensureBlankLine();
                if (this._extraComments) {
                    this.emitMark("Null checks and conversion");
                }
                this.emitExtraComments("NSJSONSerialization uses NSNull instead of nil, but we want nil in our models.");
                this.emitMultiline("id NSNullify(id _Nullable nullable) {\n    BOOL isNull = nullable == nil || [nullable isEqualTo:[NSNull null]];\n    return isNull ? [NSNull null] : nullable;\n}\n\nid NSNullMap(id _Nullable nullable,  id (^f)(id element)) {\n    BOOL isNull = [NSNullify(nullable) isEqualTo:[NSNull null]];\n    return isNull ? [NSNull null] : f(nullable);\n}");
                this.ensureBlankLine();
                this.emitLine("#define NSNullOrNil(x) [NSNullify(x) isEqualTo:[NSNull null]]");
                this.ensureBlankLine();
                this.emitDictionaryAndArrayExtensions();
                // We wouldn't need to emit these private iterfaces if we emitted implementations in forward-order
                // but the code is more readable and explicit if we do this.
                if (this._extraComments) {
                    this.emitMark("Private model interfaces");
                }
                this.forEachNamedType("leading-and-interposing", this.emitPrivateClassInterface, function () { return null; }, function () { return null; });
                this.emitMark("JSON serialization implementations");
                this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitTopLevelFunctions(t, n); });
            }
            this.forEachNamedType("leading-and-interposing", this.emitClassImplementation, function () { return null; }, function () { return null; });
            if (!this._justTypes) {
                this.ensureBlankLine();
                this.emitLine("NS_ASSUME_NONNULL_END");
            }
        }
    };
    return ObjectiveCRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
